/*
	NSG32.C
	gdi32 wrapper for NScripter
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include "nsg32.h"


long
calclinebytes(long w, long bits)
{
	long	l;
	
	if (bits == 15) bits = 16;
	l = (w * bits + 7) >> 3;
	l = (l + 3) & (~(3UL));
	return l;
}

int
get_associated_dib(HDC hdc, DIBSECTION *sct)
{
#if 0
	HBITMAP	hb;
	int		siz = 0;

	ZeroMemory(sct, sizeof(DIBSECTION));
	hb = GetCurrentObject(hdc, OBJ_BITMAP);
	if (hb) {
		siz = GetObject(hb, 0, NULL);
		GetObject(hb, siz, sct);
	}
	
#else
	HBITMAP	hb, hbp;
	int		siz;
	
	ZeroMemory(sct, sizeof(DIBSECTION));
	hb = CreateCompatibleBitmap(hdc, 1, 1);
	hbp = (HBITMAP)SelectObject(hdc, hb);
	siz = GetObject(hbp, 0, NULL);
	if (siz == sizeof(BITMAP) || siz == sizeof(DIBSECTION)) {
		GetObject(hbp, sizeof(DIBSECTION), sct);
		// for odin...
		if (siz == sizeof(BITMAP) && (sct->dsBmih.biBitCount == 8 || sct->dsBmih.biBitCount == 24)) siz = sizeof(DIBSECTION);
	}
	SelectObject(hdc, hbp);
	DeleteObject(hb);
#endif
	return siz;
}

static
BOOL
bltsection(BYTE *destmem, BITMAPINFOHEADER *destbi, int xdest, int ydest, int w, int h, CONST BYTE *srcmem, CONST BITMAPINFOHEADER *srcbi, int xsrc, int ysrc, DWORD rop)
{
	long	cbdl, cbsl;
	BYTE	*d, *s;
	
	if (destbi->biBitCount != 24 || srcbi->biBitCount != 24) return FALSE;
	
	if (xsrc < 0) { xdest -= xsrc; w += xsrc; xsrc = 0; }
	if (xdest < 0) { xsrc -= xdest; w += xdest; xdest = 0; }
	if (xsrc + w > srcbi->biWidth) { w = srcbi->biWidth - xsrc; }
	if (xdest + w > destbi->biWidth) { w = destbi->biWidth - xdest; }
	
	if (ysrc < 0) { ydest -= ysrc; h += ysrc; ysrc = 0; }
	if (ydest < 0) { ysrc -= ydest; h += ydest; ydest = 0; }
	if (ysrc + h > srcbi->biHeight) { h = srcbi->biHeight - ysrc; }
	if (ydest + h > destbi->biHeight) { h = destbi->biHeight - ydest; }
	
	if (w <= 0 || h <= 0) return TRUE;
	
	cbsl = calclinebytes(srcbi->biWidth, srcbi->biBitCount);
	cbdl = calclinebytes(destbi->biWidth, destbi->biBitCount);
	while(h > 0) {
		--h;
		d = (BYTE *)destmem + cbdl * (destbi->biHeight - ydest - h - 1);
		s = (BYTE *)srcmem + cbsl * (srcbi->biHeight - ysrc - h - 1);
		memcpy(d + xdest * 3, s + xsrc * 3, w * 3);
	}
	
	return TRUE;
}


MY_EXPORT
BOOL  MYEXPORT
expBitBlt(HDC hdcdest, int xdest, int ydest, int width, int height,
		HDC hdcsrc, int xsrc, int ysrc, DWORD rop)
{
	BOOL	rc;
	int			nbs, nbd;
	DIBSECTION	dss, dsd;
	
	nbs = get_associated_dib(hdcsrc, &dss);
	nbd = get_associated_dib(hdcdest, &dsd);
	
	if (nbs == sizeof(DIBSECTION) && nbd == sizeof(DIBSECTION)) {
		rc = bltsection((BYTE *)(dsd.dsBm.bmBits), &(dsd.dsBmih),
						xdest, ydest, width, height,
						(BYTE *)(dss.dsBm.bmBits), &(dss.dsBmih),
						xsrc, ysrc, rop);
	}
	else rc = BitBlt(hdcdest, xdest, ydest, width, height, hdcsrc, xsrc, ysrc, rop);
	
	
	return rc;
}

MY_EXPORT
HDC  MYEXPORT expCreateCompatibleDC(HDC hdc)
{
	HDC	hdcdup;
	
	hdcdup = CreateCompatibleDC(hdc);
	
	return hdcdup;
}

MY_EXPORT
HBITMAP  MYEXPORT
expCreateDIBSection(HDC hdc, CONST BITMAPINFO *bi, UINT usage,
							VOID **bits, HANDLE hfile, DWORD offset)
{
	HBITMAP		hdib;
	
	hdib = CreateDIBSection(hdc, bi, usage, bits, hfile, offset);
	return hdib;
}

MY_EXPORT
HFONT  MYEXPORT
expCreateFontA(int height, int width, int escape, int orient, int weight, DWORD italic, DWORD under, DWORD strike, DWORD charset, DWORD outprec, DWORD clipprec, DWORD quality, DWORD pitch, LPCSTR face)
{
	HFONT	hf;
	
	weight = 0;		// ꂪȂƁAOS/2 ł͂ȂߎSȂƂ 
	
	// outprec = OUT_DEFAULT_PRECIS;
	// clipprec = CLIP_DEFAULT_PRECIS;
	// quality = PROOF_QUALITY;
	// pitch = pitch & 0x00fc;
	// charset = SHIFTJIS_CHARSET;
	
	hf = CreateFontA(height, width,
				escape, orient,
				weight,
				italic, under, strike,
				charset,
				outprec,
				clipprec,
				quality,
				pitch,
				face
				);

	return hf;
}

MY_EXPORT
HFONT MYEXPORT
expCreateFontIndirectA(const LOGFONTA *lf)
{
	HFONT	hf;
	
	hf = CreateFontIndirectA(lf);
	return hf;
}

MY_EXPORT
HBRUSH  MYEXPORT
expCreateSolidBrush(COLORREF col)
{
	HBRUSH	hbr;
	
	hbr = CreateSolidBrush(col);
	return hbr;
}

MY_EXPORT
BOOL  MYEXPORT
expDeleteDC(HDC hdc)
{
	BOOL	rc;
	
	rc = DeleteDC(hdc);
	return rc;
}

MY_EXPORT
BOOL  MYEXPORT
expDeleteObject(HGDIOBJ hobj)
{
	BOOL	rc;
	
	rc = DeleteObject(hobj);
	return rc;
}

MY_EXPORT
int  MYEXPORT
expEnumFontFamiliesExA(HDC hdc, LPLOGFONTA lf, FONTENUMPROCA fproc, LPARAM lparam, DWORD flags)
{
	int		rc;
	
	rc = EnumFontFamiliesExA(hdc, lf, fproc, lparam, flags);
	return rc;
}

MY_EXPORT
int  MYEXPORT
expGetDeviceCaps(HDC hdc, int n)
{
	int		rc;
	
	rc = GetDeviceCaps(hdc, n);
	return rc;
}

MY_EXPORT
HGDIOBJ  MYEXPORT
expGetStockObject(int n)
{
	HGDIOBJ	hobj;
	
	hobj = GetStockObject(n);
	return hobj;
}

MY_EXPORT
HGDIOBJ  MYEXPORT
expSelectObject(HDC hdc, HGDIOBJ hobj)
{
	HGDIOBJ	hprev;
	
	hprev = SelectObject(hdc, hobj);
	return hprev;
}

MY_EXPORT
COLORREF  MYEXPORT
expSetBkColor(HDC hdc, COLORREF c)
{
	COLORREF	cprev;
	
	cprev = SetBkColor(hdc, c);
	return cprev;
}

MY_EXPORT
int  MYEXPORT
expSetBkMode(HDC hdc, int mode)
{
	int		prev;
	
	prev = SetBkMode(hdc, mode);
	return prev;
}

MY_EXPORT
COLORREF  MYEXPORT
expSetTextColor(HDC hdc, COLORREF c)
{
	COLORREF	cprev;
	
	cprev = SetTextColor(hdc, c);
	return cprev;
}

MY_EXPORT
BOOL  MYEXPORT
expStretchBlt(HDC hdcdest, int xdest, int ydest, int wdest, int hdest,
			HDC hdcsrc, int xsrc, int ysrc, int wsrc, int hsrc, DWORD rop)
{
	BOOL	rc;

	rc = StretchBlt(hdcdest, xdest, ydest, wdest, hdest,
					hdcsrc, xsrc, ysrc, wsrc, hsrc, rop);
	return rc;
}


MY_EXPORT
BOOL  MYEXPORT
expTextOutA(HDC hdc, int x, int y, LPCSTR str, int length)
{
	BOOL		rc;
	int			siz;
	DIBSECTION	sct;
	
	siz = get_associated_dib(hdc, &sct);
	if (siz == sizeof(DIBSECTION) && sct.dsBmih.biBitCount == 24) {
		BITMAPINFOHEADER *bi = (BITMAPINFOHEADER *)&(sct.dsBmih);
		HDC		hdc2 = CreateCompatibleDC(hdc);
		HBITMAP	hbmp, hbprev;
		HFONT	hf, hfprev;
		
		hf = (HFONT)SelectObject(hdc, GetStockObject(SYSTEM_FONT));
		
		hfprev = (HFONT)SelectObject(hdc2, hf);
		SetTextColor(hdc2, GetTextColor(hdc));
		SetBkColor(hdc2, GetBkColor(hdc));
		SetBkMode(hdc2, GetBkMode(hdc));
		
		bi->biSize = sizeof(BITMAPINFOHEADER);
		hbmp = CreateDIBitmap(hdc2, bi, CBM_INIT, sct.dsBm.bmBits, (BITMAPINFO *)bi, DIB_RGB_COLORS);
		if (hbmp) {
			hbprev = (HBITMAP)SelectObject(hdc2, hbmp);
			rc = TextOutA(hdc2, x, y, str, length);
			SelectObject(hdc2, hbprev);
			GetDIBits(hdc2, hbmp, 0, bi->biHeight, sct.dsBm.bmBits, (BITMAPINFO *)bi ,DIB_RGB_COLORS);
			DeleteObject(hbmp);
		}
		
		SelectObject(hdc2, hfprev);
		SelectObject(hdc, hf);
		DeleteDC(hdc2);
	}
	else rc = TextOutA(hdc, x, y, str, length);
	
	
	return rc;
}



BOOL
OnDLLInit()
{
	return TRUE;
}


BOOL
OnDLLTerm()
{
	return TRUE;
}

int APIENTRY
#ifdef __WATCOMC__
LibMain
#else
DllMain
#endif
(HINSTANCE hi, DWORD fdwReason, LPVOID reserved)
{
	int		rc;
	
	switch(fdwReason) {
		case DLL_PROCESS_ATTACH:
			rc = OnDLLInit();
			break;
		case DLL_PROCESS_DETACH:
			rc = OnDLLTerm();
		default:
			rc = TRUE;
	}
	return rc;
}
